from pathlib import Path
import rhasspynlu
from rhasspyasr_kaldi.transcribe import KaldiCommandLineTranscriber
from utils import kaldi_path
from rhasspynlu.fsticuffs import recognize


class Transcriber:
    def __init__(self,
                 path: Path
                 ):
        with open(path / 'intent_graph.pickle.gz', 'rb') as file:
            self.graph = rhasspynlu.gzip_pickle_to_graph(file)

        self.transcriber = KaldiCommandLineTranscriber(
            model_type = 'nnet3',
            model_dir = path / 'kaldi/model',
            graph_dir = path / 'kaldi/model/graph',
            port_num = 50847,
            kaldi_dir = kaldi_path,
            kaldi_args = None
        )

    def transcribe(self, wav: bytes):
        # Parameters 16000, 2, 1 are actually not correct for test items.
        # It doesn't seem that in Rhasspy these parameters are possible to change for Kaldi transcriber
        # They might not matter. If they do, some workaround needs to be developed to obtain them
        transcription = self.transcriber.transcribe_stream([wav], 16000, 2, 1)

        recognition = recognize(
            tokens = transcription.text,
            graph = self.graph,
            fuzzy = True,
            stop_words = None,
            word_transform=str.lower,
            converters=None,
            extra_converters={},
        )

        from copy import copy
        js_recognition = []
        for rec in recognition:
            js_rec = copy(rec.__dict__)
            js_rec['entities'] = [e.__dict__ for e in js_rec['entities']]
            js_recognition.append(js_rec)

        js_transcription = copy(transcription.__dict__)
        js_transcription['tokens'] = [t.__dict__ for t in js_transcription['tokens']]

        result = dict(fsticuffs=js_recognition, kaldi=js_transcription)
        return result




import sys
import subprocess
import argparse

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('-n', '--notebook', action='store_true')
    parser.add_argument('-r', '--rhasspy', action='store_true')
    parser.add_argument('-l', '--language')

    args = parser.parse_args()
    print(f"Running with arguments\n{args}")

    if args.notebook:
        subprocess.call([sys.executable, '-m', 'notebook', '--allow-root', '--port', '8899', '--ip', '0.0.0.0',"--NotebookApp.token=''"], cwd='/repo')
        exit(0)

    if args.rhasspy:
        subprocess.call(["bash","/usr/lib/rhasspy/bin/rhasspy-voltron",'--user-profiles', '/profiles', '--profile', args.language])
        exit(0)

    from utils import load_rhasspy_modules
    load_rhasspy_modules()
    from server import RhasspyKaldiServer
    RhasspyKaldiServer()()


from brainbox.deciders.images.video_to_images.controller import VideoToImagesController
from brainbox.utils.compatibility import *

if __name__ == '__main__':
    controller = VideoToImagesController()
    #resolve_dependencies(controller)
    controller.install()
    controller.self_test()
    #test_on_arm64(controller)







from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from ..core import JobForPlanner, OperatorStateForPlanner, OperatorLogHandle, OperatorLogItem
from .planner_action import IPlannerAction





@dataclass
class PlannerArguments:
    non_finished_tasks: tuple[JobForPlanner,...]
    deciders: tuple[OperatorStateForPlanner,...]
    log_handler: OperatorLogHandle = field(default_factory=lambda:OperatorLogHandle(None,OperatorLogItem.Level.Planer,None))


class IPlanner(ABC):
    @abstractmethod
    def plan(self, arguments: PlannerArguments) -> list[IPlannerAction]:
        pass


from typing import Iterable
from unittest import TestCase
from ....framework import (
    RunConfiguration, TestReport, BrainboxImageBuilder, IImageBuilder, DockerWebServiceController,
    BrainBoxApi, BrainBoxTask, FileIO, INotebookableController, IModelDownloadingController, DownloadableModel
)
from .settings import EspeakPhonemizerSettings
from pathlib import Path


class EspeakPhonemizerController(
    DockerWebServiceController[EspeakPhonemizerSettings],
    INotebookableController,
):
    def get_image_builder(self) -> IImageBuilder|None:
        return BrainboxImageBuilder(
            Path(__file__).parent,
            '3.8.20',
            ('libespeak-ng1',),
            allow_arm64=True
        )

    def get_service_run_configuration(self, parameter: str|None) -> RunConfiguration:
        if parameter is not None:
            raise ValueError(f"`parameter` must be None for {self.get_name()}")
        return RunConfiguration(
            publish_ports={self.connection_settings.port:8080},
        )

    def get_notebook_configuration(self) -> RunConfiguration|None:
        return self.get_service_run_configuration(None).as_notebook_service()

    def get_default_settings(self):
        return EspeakPhonemizerSettings()

    def create_api(self):
        from .api import EspeakPhonemizer
        return EspeakPhonemizer()

    def _self_test_internal(self, api: BrainBoxApi, tc: TestCase) -> Iterable:
        from .api import EspeakPhonemizer

        api.execute(BrainBoxTask.call(EspeakPhonemizer).phonemize("This is some text"))

        yield (
            TestReport
            .last_call(api)
            .href('en-pho')
            .with_comment("Phonemizes one english line")
        )

        api.execute(BrainBoxTask.call(EspeakPhonemizer).phonemize(["This is some text", "This is the second line"]))

        yield (
            TestReport
            .last_call(api)
            .href('en-pho-2')
            .with_comment("Phonemizes several english lines")
        )

        api.execute(BrainBoxTask.call(EspeakPhonemizer).phonemize("Keep the stress marks", stress=True))
        yield (
            TestReport
            .last_call(api)
            .href('stress')
            .with_comment("Phonemizes keeping the stress marks")
        )

        api.execute(BrainBoxTask.call(EspeakPhonemizer).phonemize('Das is ein Satz', language='de'))
        yield (
            TestReport
                .last_call(api)
                .href('de')
                .with_comment("Phonemizes German")
        )





from .template import TemplateWorkflow
from dataclasses import dataclass, field
import numpy as np
from .....framework import BrainBox
import pprint

@dataclass
class TextToImage(TemplateWorkflow):
    prompt: str = field(metadata=TemplateWorkflow.meta('text', 'CLIP Text Encode (Positive)'))
    negative_prompt: str = field(default='', metadata=TemplateWorkflow.meta('text', 'CLIP Text Encode (Negative)'))
    width: int = 512
    height: int = 512
    batch_size: int = 1
    seed: int = field(default_factory=lambda:np.random.randint(0, 1000000))
    steps: int = 20
    cfg: float = 8
    sampler_name: str = 'euler'
    scheduler: str = 'normal'
    denoise: float = 1.0
    model: str = field(default='v1-5-pruned-emaonly.ckpt', metadata=TemplateWorkflow.meta('ckpt_name', is_model_with_priority=0))
    filename_prefix: str = field(default='', metadata=TemplateWorkflow.meta(maps_to_job_id=True))
    lora_01: str|None = field(default="None", metadata=TemplateWorkflow.meta(is_model_with_priority=1))
    strength_01: float = 1.0
    lora_02: str|None = field(default="None", metadata=TemplateWorkflow.meta(is_model_with_priority=1))
    strength_02: float = 1.0
    lora_03: str|None = field(default="None", metadata=TemplateWorkflow.meta(is_model_with_priority=1))
    strength_03: float = 1.0
    lora_04: str|None = field(default="None", metadata=TemplateWorkflow.meta(is_model_with_priority=1))
    strength_04: float = 1.0

    def __post_init__(self):
        for i in range(1,5):
            name = f'lora_0{i}'
            if getattr(self, name) is None:
                setattr(self, name, "None")






from typing import *
from .file import File
from pathlib import Path
from io import BytesIO
from dataclasses import dataclass
from uuid import uuid4


class FileLike:
    Type = str|Path|bytes|File|BytesIO

    @dataclass
    class AlreadyStream:
        stream: Any


    def __init__(self, file: str|Path|bytes|BytesIO, cache_folder: None|Path):
        self.file = file
        self.cache_folder = cache_folder
        self.stream = None

    @staticmethod
    def get_name(file: 'FileLike.Type', raise_if_no_name: bool = False):
        if isinstance(file, str):
            return file
        elif isinstance(file, Path):
            return file.name
        elif isinstance(file, File):
            return file.name
        else:
            if raise_if_no_name:
                raise ValueError("This FileLike must define a name, but doesn't")
            return str(uuid4())

    @staticmethod
    def get_name_and_extension(file: 'FileLike.Type'):
        parts = FileLike.get_name(file).split('.')
        if len(parts) == 1:
            return parts[0], ''
        return '.'.join(parts[:-1]), '.'+parts[-1]

    def get_path(self, extension_if_caching: str|None = None):
        if isinstance(self.file, Path):
            return self.file
        elif isinstance(self.file, str):
            if self.cache_folder is None:
                return self.file
            else:
                return self.cache_folder/self.file
        else:
            name = str(uuid4())
            if extension_if_caching is not None:
                name += '.' + extension_if_caching
            path = self.cache_folder/name
            with self as stream:
                data = stream.read()
                with open(path,'wb') as out_stream:
                    out_stream.write(data)
            return path

    def __enter__(self):
        if isinstance(self.file, Path):
            self.stream = open(self.file, 'rb')
        elif isinstance(self.file, str):
            if self.cache_folder is None:
                self.stream = open(self.file, 'rb')
            else:
                self.stream = open(self.cache_folder/self.file, 'rb')
        elif isinstance(self.file, bytes):
            self.stream = BytesIO(self.file)
        elif isinstance(self.file, File):
            if self.file.content is not None:
                self.stream = BytesIO(self.file.content)
            elif self.cache_folder is not None:
                self.stream = open(self.cache_folder / self.file.name, 'rb')
            else:
                self.stream = open(self.file.name, 'rb')
        elif isinstance(self.file, FileLike.AlreadyStream):
            self.stream = self.file.stream
        else:
            raise ValueError(f'File should be path, str (with path) or bytes, but was: {self.file}')
        return self.stream

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stream.close()


from .image_source import IImageSource
from .. import IExecutor


class RemotePublicImageSource(IImageSource):
    def __init__(self, remote_image_name, container_name):
        self.image_name = remote_image_name
        self.container_name = container_name

    def get_image_name(self):
        return self.image_name

    def push(self, executor: IExecutor):
        pass

    def pull(self, executor: IExecutor):
        executor.execute(['docker', 'pull', self.image_name])

    def get_container_name(self):
        return self.container_name





import argparse
import json

from functions import Functions
from settings import TrainingSettings
from pathlib import Path

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('-d', '--dataset')
    parser.add_argument('-c', '--export', action='store_true')
    parser.add_argument('-f', '--working-folder', default='/resources')

    args = parser.parse_args()
    print(f"Running with arguments\n{args}")

    if args.export:
        export_path = Path(args.working_folder)
        f = Functions(args.dataset, args.working_folder, None)
        print(f'EXPORTING DATASET {f.dataset}')
        f.export_model()
        exit(0)

    settings_path = Path(args.working_folder)/'trainings'/args.dataset/'settings.json'
    with open(settings_path,'r') as stream:
        settings_js = json.load(stream)
    settings = TrainingSettings(**settings_js)
    print("Training settings:")
    print(settings)

    f = Functions(args.dataset, args.working_folder, settings)
    print(f'DATASET {f.dataset}')
    if not settings.continue_existing:
        print('PREPROCESSING')
        f.preprocess()
    print('TRAINING')
    f.train()






from .architecture import AbstractRepresentationStep, RepresentationData
import numpy as np
from dataclasses import dataclass


def get_random_weighed_element(elements) -> int:
    csum = []
    for element in elements:
        if len(csum) == 0:
            csum.append(element)
        else:
            csum.append(csum[-1]+element)
    rnd = np.random.rand()*csum[-1]
    for i, cs in enumerate(csum):
        if cs > rnd:
            return i
    return len(csum) - 1


@dataclass
class ProbabilisticEqualRepresentationStep(AbstractRepresentationStep):
    magnitude: float = 1

    def select(self, data: list[RepresentationData]) -> None:
        weights = []
        for item in data:
            w = 1 / ( 1+ item.category_size) ** self.magnitude
            weights.append(w)

        seen = set()
        output = []
        for i in range(self.count):
            while True:
                idx = get_random_weighed_element(weights)
                if idx in seen:
                    continue
                data[idx].selected = True
                weights[idx] = 0
                break







import os.path
from typing import *
from dataclasses import dataclass, field
from ....common import File
from datetime import datetime

@dataclass
class TestReportItem:
    decider: str
    method: str | None
    parameter: str | None
    arguments: dict
    result: Any
    accepted_timestamp: datetime
    finished_timestamp: datetime

    href: str|None = None
    comment: str | None = None
    result_is_file: bool = False
    result_is_array_of_files: bool = False
    uploaded_files: dict[str, File] = field(default_factory=dict)
    used_resources: dict[str, File] = field(default_factory=dict)



class TestReportItemHolder:
    def __init__(self, api, item: TestReportItem):
        self.item = item
        self.api = api


    def with_comment(self, comment: str):
        self.item.comment = comment
        return self

    def result_is_file(self, kind: File.Kind | None = None):
        self.item.result = self.api.open_file(self.item.result)
        if kind is not None:
            if self.item.result.kind != kind:
                raise ValueError(f"Expected kind {kind}, but was {self.item.result.kind}")
        self.item.result_is_file = True
        return self

    def result_is_array_of_files(self, kind: File.Kind | None = None):
        self.item.result = [self.api.open_file(r) for r in self.item.result]
        if kind is not None:
            for r in self.item.result:
                if r.kind != kind:
                    raise ValueError(f"Expected kind {kind}, but was {r.kind}")
        self.item.result_is_array_of_files = True
        return self

    def with_uploaded_file(self, fname: str, file: File):
        self.item.uploaded_files[fname] = file
        return self

    def with_resources(self, decider: Type, path: str):
        lst = self.api.controller_api.list_resources(decider, path)
        for file in lst:
            self.item.used_resources[file] = self.api.controller_api.download_resource(decider, file)
        return self

    def href(self, href: str):
        self.item.href = href
        return self


from dataclasses import dataclass
from ....framework import ConnectionSettings
from .model import FasterWhisperModel


@dataclass
class FasterWhisperSettings:
    connection: ConnectionSettings = ConnectionSettings(20102)

    models_to_download: tuple[FasterWhisperModel,...] = (

    )





import json

class HTML:
    @staticmethod
    def escape_code(code):
        code = (
            code
            .replace("<", "&lt;")
            .replace(">", "&gt;")
            .replace('&', "&amp;")
            .replace(" ", "&nbsp")
            .replace("\n", "<br>")
        )
        return code

    @staticmethod
    def button(address, title, parameters, control_parameters: dict[str, str] | None = None, active=True):
        btn = f'<button '
        if not active:
            btn += 'disabled '  # Add the 'disabled' attribute if the button is inactive
        btn += f'''onClick="makeRequest('{address}','''
        btn += json.dumps(parameters).replace('"', "'")
        btn += ','
        if control_parameters is None:
            btn += '{}'
        else:
            btn += json.dumps(control_parameters).replace('"', "'")
        btn += f')">{title}</button>'
        return btn

    @staticmethod
    def page(title: str, content:str|list[str]):
        if isinstance(content, str):
            pass
        else:
            content = "\n".join(content)
        return HTML.header(title)+content+HTML.footer()

    @staticmethod
    def header(title):
        return f'''
        <!DOCTYPE html>
        <html lang="en">
        <head>
            <title>{title}</title>
            <meta charset="UTF-8">
            <meta name="viewport" content="width=device-width, initial-scale=1.0">
            </style>
        </head>
        <body>
        
        '''

    @staticmethod
    def footer():
        return '''
        <script>
        function makeRequest(address, jsonParameters, inputs) {
            console.log("Sending")
            // Disable all buttons on the page
            const buttons = document.querySelectorAll('button');
            buttons.forEach(button => button.disabled = true);
            
            for (const [key, elementId] of Object.entries(inputs)) {
                const inputElement = document.getElementById(elementId);
                if (inputElement) {
                    jsonParameters[key] = inputElement.value;
                }
            }
            
            console.log(jsonParameters)

            // Send POST request
            fetch(address, {
                method: 'POST',
                headers: {
                    'Content-Type': 'application/json'
                },
                body: JSON.stringify({"arguments":jsonParameters})
            })
            .then(response => {
                //buttons.forEach(button => button.disabled = false);
                window.location.reload();
            })
            .catch(error => {
                console.error('There was a problem with the fetch operation:', error);
                // Optionally re-enable buttons if there is an error
                buttons.forEach(button => button.disabled = false);
            });
        }
        </script>
        </body>
        </html>
        '''


from unittest import TestCase
from ....framework import TestReport, File, BrainBoxApi, BrainBoxTask
from .api import RhasspyKaldi
from pathlib import Path

def english(api: BrainBoxApi, tc: TestCase):
    yield TestReport.section("English model")
    model_name = 'test_english'

    sentences = '''
[timer]
minutes = one | two | three | four | five | six | seven | eight | nine | ten
Set the timer for <minutes> {minutes} minutes
        
[time]
What time is it
    '''.strip()
    api.execute(BrainBoxTask.call(RhasspyKaldi).train(model_name, 'en', sentences))
    yield TestReport.last_call(api).href('train-en').with_comment("Training the model")


    file = File.read(Path(__file__).parent / 'files/timer.wav')
    result = api.execute(BrainBoxTask.call(RhasspyKaldi).transcribe(file, model_name))
    yield TestReport.last_call(api).href('inference-en').with_comment("Inference")
    tc.assertEqual('five', result['fsticuffs'][0]['entities'][0]['raw_value'])
    tc.assertEqual('timer', result['fsticuffs'][0]['intent']['name'])





def german(api: BrainBoxApi, tc: TestCase):
    yield TestReport.section("German model")
    model_name = 'test_german'

    sentences = '''
[timer_de]
_minutes = zwei|drei|vier|fünf
Stelle den Timer auf <_minutes>{_minutes} Minuten    
    '''.strip()

    api.execute(BrainBoxTask.call(RhasspyKaldi).train(model_name,'de',sentences))
    yield TestReport.last_call(api).href('train-de').with_comment("Training the model")

    file = File.read(Path(__file__).parent / 'files/timer-de.wav')
    result = api.execute(BrainBoxTask.call(RhasspyKaldi).transcribe(file, model_name))
    yield TestReport.last_call(api).href('inference-de').with_comment("Inference with a trained model")
    tc.assertEqual('fünf', result['fsticuffs'][0]['entities'][0]['raw_value'])
    tc.assertEqual('timer_de', result['fsticuffs'][0]['intent']['name'])



def english_custom(api: BrainBoxApi, tc: TestCase):
    yield TestReport.section("English model with custom pronunciation")
    model_name = 'test_english_custom_words'

    sentences = '''
[order]
_order = Shi|Ukha|Borsh
I order <_order>{_order}    
        '''.strip()

    result = api.execute(BrainBoxTask.call(RhasspyKaldi).train(model_name, 'en', sentences))
    yield (TestReport
            .last_call(api)
            .href('train-custom-1')
            .with_comment("Training with words that are missing from the dictionary. The training returns the list of words in question.")
           )
    tc.assertDictEqual(
        dict(borsh="b 'O r S", ukha="j u k V"),
        result['unknown_words']
    )

    api.execute(BrainBoxTask.call(RhasspyKaldi).phonemes('en'))
    yield TestReport.last_call(api).href('transcribtions').with_comment("Transcription can be provided for these words. The symbols for this transcription are rather peculiar, and are returned by this endpoint.")

    custom_words = {
        'shi' : "S i",
        'ukha': "u h 'A",
        'borsh': 'b o r S'
    }
    result = api.execute(BrainBoxTask.call(RhasspyKaldi).train(model_name, 'en', sentences, custom_words))

    yield TestReport.last_call(api).href('train-custom-2').with_comment("The training with the transcription of custom words specified")
    tc.assertEqual(0, len(result['unknown_words']))

    file = File.read(Path(__file__).parent / 'files/ukha.wav')
    result = api.execute(BrainBoxTask.call(RhasspyKaldi).transcribe(file, model_name))
    yield TestReport.last_call(api).href('inference-custom').with_comment("The inference with custom pronunciation")
    tc.assertEqual('ukha', result['fsticuffs'][0]['entities'][0]['raw_value'])
    tc.assertEqual('order', result['fsticuffs'][0]['intent']['name'])







import time
import traceback

from ..core import Core, OperatorStateForPlanner
from .check_ready_action import CheckReadyAction
from .remove_incorrect_action import RemoveIncorrectJobsAction
#from .tasks_for_planner_sync_with_tracking import TaskForPlannerSyncWithTracking
from .tasks_for_planner_sync_with_last_updated import TaskForPlannerSyncWithLastUpdated
from .task_status_updater import TasksStatusUpdater
from .cancel_action import CancelAction
from ..planner import IPlanner, PlannerArguments
from ...job_processing.planner.stop_action import StopCommand
from .dirty_exit_fix import DirtyExitFix

class MainLoop:
    def __init__(self,
                 core: Core,
                 planner: IPlanner,
                 stop_containers_at_termination: bool = True
                 ):
        self.core = core
        self.planner = planner
        self._terminate_request: bool = False
        self.stop_containers_at_termination = stop_containers_at_termination

        
    def run(self):
        DirtyExitFix().apply(self.core)
        while True:
            if self._terminate_request:
                break
            try:
                self.one_iteration()
            except:
                message = traceback.format_exc()
                self.core.operator_log.core().error(traceback.format_exc())
                print(message)
            time.sleep(0.1)
        if self.stop_containers_at_termination:
            for state in list(self.core.operator_states.values()):
                try:
                    StopCommand(state.controller_run_instance_id).apply(self.core)
                except:
                    pass



    def terminate(self):
        self._terminate_request = True



    def cancel(self, job_id: str|None = None, batch_id: str|None = None):
        CancelAction(job_id, batch_id).apply(self.core)


    def one_iteration(self):
        RemoveIncorrectJobsAction().apply(self.core)
        TasksStatusUpdater().apply(self.core)
        CheckReadyAction().apply(self.core)
        TaskForPlannerSyncWithLastUpdated().apply(self.core)

        for action in self._run_planner():
            action.apply(self.core)


    def _run_planner(self):
        states_for_planner = tuple(
            OperatorStateForPlanner(instance_id, state.key, state.busy, state.not_busy_since)
            for instance_id, state
            in self.core.operator_states.items()
        )

        planner_arguments = PlannerArguments(
            tuple(job for job in self.core.jobs_for_planner),
            states_for_planner,
            self.core.operator_log.planer()
        )

        return self.planner.plan(planner_arguments)





from brainbox.deciders.voice_analysis.rhasspy_kaldi import RhasspyKaldi

if __name__ == '__main__':
    installer = RhasspyKaldi.Controller()
    installer.uninstall(True)
    installer.install()
    installer.self_test()
    


import traceback
from dataclasses import dataclass, field
from ..parsers import IParser
from .applicator import IApplicator
from pathlib import Path
from yo_fluq import *
from .object_to_task import IObjectConverter, PromptBasedObjectConverter
from ..cache import JsonCacheManager, ICacheManager
from foundation_kaia.prompters import Address

T = TypeVar('T')

@dataclass
class BrainBoxMappingError:
    brainbox_error: str|None = None
    parse_error: str|None = None
    not_found_in_cache: bool = False



@dataclass
class BrainBoxMappingData(Generic[T]):
    mapping: 'BrainBoxMapping[T]'
    source: tuple[T,...]
    errors: tuple[BrainBoxMappingError,...]
    cache: Path | None
    tasks: tuple[IObjectConverter.Output,...]|None = None
    is_task_cached: tuple[bool,...]|None = None
    llm_answer: tuple[str|None,...]|None = None
    parsed_llm_answer: tuple[Any,...]|None = None
    result: tuple[T,...]|None = None
    _cache_manager: ICacheManager[str] = field(default_factory=ICacheManager)

    def __post_init__(self):
        if self.cache is not None:
            self._cache_manager = JsonCacheManager(self.cache)

    def build_task_and_tags(self):
        tasks = []
        is_cached = []
        for index, source in enumerate(self.source):
            task = self.mapping.object_to_task.convert(source)
            is_cached.append(self._cache_manager.contains(task.cache_key))
            task.tags['index'] = index
            tasks.append(task)
        self.tasks = tuple(tasks)
        self.is_task_cached = tuple(is_cached)


    def build_collector_task(self):
        if self.tasks is None:
            self.build_task_and_tags()
        try:
            from brainbox import BrainBox
            from brainbox.deciders import Ollama, Collector
        except ImportError as ex:
            raise "Install kaia-brainbox to use the automatical task building" from ex

        builder = Collector.TaskBuilder()
        for task, is_cached in zip(self.tasks, self.is_task_cached):
            if is_cached:
                continue
            builder.append(task.task, task.tags)
        return builder.to_collector_pack('to_array')


    def _consume_results(self, result):
        index_to_result = {}
        for item in result:
            index = item['tags']['index']
            if item['error'] is not None:
                self.errors[index].llm_error = item['error']
                index_to_result[index] = None
            else:
                index_to_result[index] = item['result']

        llm_answer = [None for _ in self.source]
        for index, task in enumerate(self.tasks):
            if index in index_to_result:
                result = index_to_result[index]
                llm_answer[index] = result
                self._cache_manager.store(task.cache_key, result)
            else:
                result = self._cache_manager.get(task.cache_key)
                if result is not None:
                    llm_answer[index] = result
                else:
                    self.errors[index].not_found_in_llm_nor_cache = True
        self.llm_answer = tuple(llm_answer)

    def _parse(self):
        parsed = []
        for index, item in enumerate(self.llm_answer):
            if item is None:
                parsed.append(None)
                continue
            if self.mapping.parser is None:
                parsed.append(item)
                continue
            try:
                value = self.mapping.parser(item)
            except:
                self.errors[index].parse_error = traceback.format_exc()
                value = None
            parsed.append(value)
        self.parsed_llm_answer = tuple(parsed)

    def _apply(self):
        application = []
        for item, parsed, task in zip(self.source, self.parsed_llm_answer, self.tasks):
            if parsed is None:
                continue
            for result in self.mapping.applicator.apply(item, parsed):
                if self.mapping.tags_address is not None:
                    self.mapping.tags_address.set(result, task.tags)
                application.append(result)
        self.result = tuple(application)


    def apply_result(self, result):
        self._consume_results(result)
        self._parse()
        self._apply()

    def execute(self, api) -> 'BrainBoxMappingData[T]':
        self.build_task_and_tags()
        task = self.build_collector_task()
        result = api.execute(task)
        self.apply_result(result)
        return self


@dataclass
class BrainBoxMapping(Generic[T]):
    object_to_task: IObjectConverter
    applicator: IApplicator
    parser: IParser|None = None
    tags_address: Address|None = None
    data: BrainBoxMappingData[T]|None = None

    def create(self, source: Iterable[T], cache: Path|None = None) -> BrainBoxMappingData[T]:
        source = tuple(source)
        errors = tuple(BrainBoxMappingError() for _ in source)
        return BrainBoxMappingData(
            self,
            source,
            errors,
            cache
        )



from .template import TemplateWorkflow
from dataclasses import dataclass, field
from .....framework import FileLike, File

@dataclass
class WD14Interrogate(TemplateWorkflow):
    filename: FileLike.Type = field(metadata=TemplateWorkflow.meta('image','Load Image', is_input_file=True))
    threshold: float = 0.35
    character_threshold: float = 0.85
    model: str = field(default='wd-v1-4-moat-tagger-v2', metadata=TemplateWorkflow.meta(is_model_with_priority=0))
    output_filename: str = field(default='', metadata=TemplateWorkflow.meta('file', maps_to_job_id=True))


    def preprocess_value(self, field_name, value):
        if field_name=='output_filename':
            return value+'.txt'
        return value

    def postprocess(self, files: list[File]):
        if len(files)!=1:
            raise ValueError(f"Exactly 1 output file is expected for WD14, but was {len(files)}")
        if files[0].kind!=File.Kind.Text:
            raise ValueError(f"Text file is expected for WD14, but was {files[0].kind}")
        return files[0].string_content






import argparse
import json
import logging
from pathlib import Path

import torch
from pytorch_lightning import Trainer
from pytorch_lightning.callbacks import ModelCheckpoint

from .vits.lightning import VitsModel

_LOGGER = logging.getLogger(__package__)


def main():
    logging.basicConfig(level=logging.DEBUG)

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--dataset-dir", required=True, help="Path to pre-processed dataset directory"
    )
    parser.add_argument(
        "--checkpoint-epochs",
        type=int,
        help="Save checkpoint every N epochs (default: 1)",
    )
    parser.add_argument(
        "--quality",
        default="medium",
        choices=("x-low", "medium", "high"),
        help="Quality/size of model (default: medium)",
    )
    parser.add_argument(
        "--resume_from_single_speaker_checkpoint",
        help="For multi-speaker models only. Converts a single-speaker checkpoint to multi-speaker and resumes training",
    )

    parser.add_argument(
        '--keep-intermediate',
        help = "Keep the intermediate checkpoint in the folder",
        action='store_true',
    )

    Trainer.add_argparse_args(parser)
    VitsModel.add_model_specific_args(parser)
    parser.add_argument("--seed", type=int, default=1234)
    args = parser.parse_args()
    _LOGGER.debug(args)

    args.dataset_dir = Path(args.dataset_dir)
    if not args.default_root_dir:
        args.default_root_dir = args.dataset_dir

    torch.backends.cudnn.benchmark = True
    torch.manual_seed(args.seed)

    config_path = args.dataset_dir / "config.json"
    dataset_path = args.dataset_dir / "dataset.jsonl"

    with open(config_path, "r", encoding="utf-8") as config_file:
        # See preprocess.py for format
        config = json.load(config_file)
        num_symbols = int(config["num_symbols"])
        num_speakers = int(config["num_speakers"])
        sample_rate = int(config["audio"]["sample_rate"])

    trainer = Trainer.from_argparse_args(args)
    if args.checkpoint_epochs is not None:
        save_top_k_argument = dict(save_top_k=-1) if args.keep_intermediate else dict()
        trainer.callbacks = [ModelCheckpoint(every_n_epochs=args.checkpoint_epochs, **save_top_k_argument)]
        _LOGGER.debug(
            "Checkpoints will be saved every %s epoch(s)", args.checkpoint_epochs
        )

    dict_args = vars(args)
    if args.quality == "x-low":
        dict_args["hidden_channels"] = 96
        dict_args["inter_channels"] = 96
        dict_args["filter_channels"] = 384
    elif args.quality == "high":
        dict_args["resblock"] = "1"
        dict_args["resblock_kernel_sizes"] = (3, 7, 11)
        dict_args["resblock_dilation_sizes"] = (
            (1, 3, 5),
            (1, 3, 5),
            (1, 3, 5),
        )
        dict_args["upsample_rates"] = (8, 8, 2, 2)
        dict_args["upsample_initial_channel"] = 512
        dict_args["upsample_kernel_sizes"] = (16, 16, 4, 4)

    model = VitsModel(
        num_symbols=num_symbols,
        num_speakers=num_speakers,
        sample_rate=sample_rate,
        dataset=[dataset_path],
        **dict_args,
    )

    if args.resume_from_single_speaker_checkpoint:
        assert (
            num_speakers > 1
        ), "--resume_from_single_speaker_checkpoint is only for multi-speaker models. Use --resume_from_checkpoint for single-speaker models."

        # Load single-speaker checkpoint
        _LOGGER.debug(
            "Resuming from single-speaker checkpoint: %s",
            args.resume_from_single_speaker_checkpoint,
        )
        model_single = VitsModel.load_from_checkpoint(
            args.resume_from_single_speaker_checkpoint,
            dataset=None,
        )
        g_dict = model_single.model_g.state_dict()
        for key in list(g_dict.keys()):
            # Remove keys that can't be copied over due to missing speaker embedding
            if (
                key.startswith("dec.cond")
                or key.startswith("dp.cond")
                or ("enc.cond_layer" in key)
            ):
                g_dict.pop(key, None)

        # Copy over the multi-speaker model, excluding keys related to the
        # speaker embedding (which is missing from the single-speaker model).
        load_state_dict(model.model_g, g_dict)
        load_state_dict(model.model_d, model_single.model_d.state_dict())
        _LOGGER.info(
            "Successfully converted single-speaker checkpoint to multi-speaker"
        )

    trainer.fit(model)


def load_state_dict(model, saved_state_dict):
    state_dict = model.state_dict()
    new_state_dict = {}

    for k, v in state_dict.items():
        if k in saved_state_dict:
            # Use saved value
            new_state_dict[k] = saved_state_dict[k]
        else:
            # Use initialized value
            _LOGGER.debug("%s is not in the checkpoint", k)
            new_state_dict[k] = v

    model.load_state_dict(new_state_dict)


# -----------------------------------------------------------------------------


if __name__ == "__main__":
    main()



from .build_context import BuildContext
from .builder_part import IBuilderPart

class User(IBuilderPart):
    def to_commands(self, context: BuildContext) -> list[str]:
        user_id = context.executor.get_machine().user_id

        result = []
        result.append(f'RUN useradd -ms /bin/bash app -u {user_id} && usermod -aG sudo app')
        result.append('USER app')
        return result



from ....framework import DockerWebServiceApi, FileLike
from .settings import InsightFaceSettings
from .controller import InsightFaceController
import requests

class InsightFace(DockerWebServiceApi[InsightFaceSettings, InsightFaceController]):
    def __init__(self, address: str|None = None):
        super().__init__(address)

    def analyze(self, image: FileLike.Type):
        with FileLike(image, self.cache_folder) as file:
            reply = requests.post(
                f'http://{self.address}/analyze',
                files=(
                    ('file', file),
                )
            )
            if reply.status_code != 200:
                raise ValueError(f"InsightFaceApi couldn't analyze\n{reply.text}")
            return reply.json()

    Controller = InsightFaceController
    Settings = InsightFaceSettings



from .docker_io_image_source import DockerIOImageSource

class HackedDockerIOImageSource(DockerIOImageSource):
    def __init__(self,
                 name: str,
                 docker_url: str,
                 registry: str,
                 login: str,
                 password: str,
                 ):
        self.registry = registry
        super().__init__(name, docker_url, login, password)

    def get_remote_name(self) -> str:
        return f'{self.docker_url}/{self.registry}:{self.name}'




from abc import ABC, abstractmethod
from ..executor import IExecutor

class IImageBuilder(ABC):
    @abstractmethod
    def build_image(self, image_name: str, exec: IExecutor) -> None:
        pass





from ..planner import IPlanner, PlannerArguments, StartCommand, AssignAction, IPlannerAction
from typing import Union

class AlwaysOnPlanner(IPlanner):
    Mode = StartCommand.Mode
    def __init__(self, mode: Union['AlwaysOnPlanner.Mode',None] = None):
        if mode is None:
            mode = AlwaysOnPlanner.Mode.FindThenStart
        self.mode = mode

    def plan(self, arguments: PlannerArguments) -> list[IPlannerAction]:
        available = {op.key:op for op in arguments.deciders}
        result = []
        for job in arguments.non_finished_tasks:
            if job.assigned:
                continue
            key = job.get_decider_instance_key()
            if key in available:
                result.append(AssignAction(job.id, available[key].instance_id, key))
                continue
            else:
                result.append(StartCommand(key, self.mode))
        return result




from typing import cast
from dataclasses import dataclass
from pathlib import Path

from ....framework import DownloadableModel, IController
from ...common import download_file


@dataclass
class PiperModel(DownloadableModel):
    name: str
    config_url: str
    url: str

    def get_local_location(self, controller: IController) -> Path:
        return controller.resource_folder()/'models'/(self.name+'.onnx')

    @classmethod
    def download(cls, models: list['DownloadableModel'], controller: IController):
        from .controller import PiperController
        controller = cast(PiperController, controller)
        path = controller.resource_folder('models')
        for model in models:
            model = cast(PiperModel, model)
            download_file(model.config_url, path/(model.name+'.onnx.json'))
            download_file(model.url, path/(model.name+'.onnx'))


from typing import TypeVar, Generic, Any, Callable
from abc import ABC, abstractmethod
from foundation_kaia.prompters import IPrompter
from dataclasses import dataclass, field
import hashlib
from brainbox import BrainBoxTask

T = TypeVar("T")

class IObjectConverter(ABC, Generic[T]):
    @dataclass
    class Output:
        task: Any
        tags: dict = field(default_factory=dict)
        cache_key: str|None = None

    @abstractmethod
    def convert(self, object: T) -> 'IObjectConverter.Output':
        pass

class PromptBasedObjectConverter(IObjectConverter[T], Generic[T]):
    def __init__(self,
                 prompter: IPrompter[T],
                 model: str,
                 decider_name: str = 'Ollama',
                 decider_method: str = 'question',
                 prompt_argument_name: str = 'prompt'
                 ):
        self.prompter = prompter
        self.model = model
        self.decider_name = decider_name
        self.decider_method = decider_method
        self.prompt_argument_name = prompt_argument_name


    def convert(self, object) -> IObjectConverter.Output:
        prompt = self.prompter(object)
        task = BrainBoxTask(
            decider=self.decider_name,
            decider_method=self.decider_method,
            arguments={self.prompt_argument_name: prompt},
            decider_parameter=self.model,
        )
        tags = dict(prompt=prompt)
        key = self.model+'/'+prompt
        key = hashlib.blake2s(key.encode(), digest_size=16).hexdigest()
        return IObjectConverter.Output(
            task,
            tags,
            key,
        )

class FunctionalObjectToTask(IObjectConverter, Generic[T]):
    def __init__(self, function: Callable[[T], IObjectConverter.Output]):
        self.function = function

    def convert(self, object: T) -> 'IObjectConverter.Output':
        return self.function(object)







from ....framework import DockerWebServiceApi, FileLike
from .settings import WD14TaggerSettings
from .controller import WD14TaggerController
from .model import WD14TaggerModel
import requests

class WD14Tagger(DockerWebServiceApi[WD14TaggerSettings, WD14TaggerController]):
    def __init__(self, address: str|None = None):
        super().__init__(address)


    def interrogate(self, image: FileLike.Type, threshold=0.35, model='wd14-vit.v2'):
        with FileLike(image, self.cache_folder) as file:
            reply = requests.post(
                f'http://{self.address}/interrogate/{model}/{threshold}',
                files=(
                    ('file', file),
                )
            )
            if reply.status_code != 200:
                raise ValueError(f"WD14Tagger couldn't interrogate for {model}\n{reply.text}")
            return reply.json()

    def tags(self, model = 'wd14-vit.v2', count=None):
        reply = requests.post(f'http://{self.address}/tags/{model}', json=dict(count=count))
        if reply.status_code != 200:
                raise ValueError(f"WD14Tagger couldn't get tags for {model}\n{reply.text}")
        return reply.json()

    Controller = WD14TaggerController
    Settings = WD14TaggerSettings
    Model = WD14TaggerModel







import os
import shutil
from typing import cast
from pathlib import Path
import zipfile

from ....framework import DownloadableModel, IController
from ...common import download_file
from dataclasses import dataclass

@dataclass
class VoskModel(DownloadableModel):
    name: str
    url: str

    def get_local_location(self, controller: IController) -> Path:
        return controller.resource_folder('models')/self.name

    @classmethod
    def download(cls, models: list['DownloadableModel'], controller: IController):
        for model in models:
            model = cast(VoskModel, model)
            zip_file_path = controller.resource_folder('models')/(model.name+'.zip')
            download_file(
                model.url,
                zip_file_path
            )
            model_path = controller.resource_folder('models')/model.name
            shutil.rmtree(model_path, ignore_errors=True)
            os.makedirs(model_path)
            with zipfile.ZipFile(zip_file_path, 'r') as zip_file:
                zip_file.extractall(model_path)
            inner_folder = os.listdir(model_path)
            if len(inner_folder) != 1:
                raise ValueError(f"after unzipping, only one folder was expected, but was {inner_folder}")
            for item in (model_path/inner_folder[0]).iterdir():
                shutil.move(str(item), str(model_path))
            os.unlink(zip_file_path)





import json

from TTS.api import TTS
import torch
from dataclasses import dataclass
from pathlib import Path

@dataclass
class TTSInfo:
    tts: TTS
    device: str
    is_custom_model: bool


def _load_custom_model(folder, filename, device):
    with open(folder / (filename + '.json')) as file:
        spec = json.load(file)
    spec['model_args']['d_vector_file'] = [str(folder / (filename + '.speakers.pth'))]
    with open(folder / (filename + '.fixed.json'), 'w') as file:
        json.dump(spec, file)
    tts = TTS(model_path=folder / filename, config_path=folder / (filename + '.fixed.json')).to(device)
    return tts

def load_tts(model_name) -> TTSInfo:
    device = "cuda" if torch.cuda.is_available() else "cpu"
    if model_name.startswith('tts_models'):
        api = TTS("")
        api.manager.output_prefix = '/resources/builtin'
        api.load_tts_model_by_name(model_name)
        tts = api.to(device)
        return TTSInfo(tts, device, False)
    if model_name.startswith('custom/'):
        folder = Path('/resources/custom')
        filename = model_name.replace('custom/','')
        tts = _load_custom_model(folder, filename, device)
        return TTSInfo(tts, device, True)
    else:
        raise ValueError('Not yet supported')


import io
import os

import flask
from tts_loader import load_tts
import traceback
from dataclasses import dataclass
from pathlib import Path

@dataclass
class ModelInfo:
    name: str
    speakers: list[str]|None
    languages: list[str]|None
    device: str
    is_custom_model: bool


class CoquiInferenceWebApp:
    def __init__(self):
        self.tts = None
        self.model: ModelInfo|None = None


    def create_app(self):
        app = flask.Flask("CoquiTTSWebApp")
        app.add_url_rule('/', view_func=self.index, methods=['GET'])
        app.add_url_rule('/load_model', view_func=self.load_model, methods=['POST'])
        app.add_url_rule('/dub', view_func=self.dub, methods=['POST'])
        app.add_url_rule('/voice_clone', view_func=self.voice_clone, methods=['POST'])
        app.add_url_rule('/get_loaded_model', view_func=self.get_loaded_model, methods=['GET'])

        return app

    def index(self):
        return f"{type(self).__name__} is running"

    def get_loaded_model(self):
        if self.model is None:
            return flask.jsonify(self.model)
        return flask.jsonify(self.model.__dict__)

    def _load_model_internal(self, model_name):
        if model_name is None:
            if self.model is None:
                raise ValueError("No model is pre-loaded and no model is specified in the request")
            else:
                return
        if self.model is not None:
            if self.model.name == model_name:
                return
        data = load_tts(model_name)
        self.tts = data.tts

        self.model = ModelInfo(
            name=model_name,
            speakers=getattr(self.tts, 'speakers', None),
            languages=getattr(self.tts, 'languages', None),
            device=data.device,
            is_custom_model=data.is_custom_model
        )

    def load_model(self):
        try:
            model_name = flask.request.json['model']
            self._load_model_internal(model_name)
            return flask.jsonify(self.model.__dict__)
        except:
            return traceback.format_exc(), 500


    def _get_language(self, language) -> str|None:
        if language is None and self.model.languages is not None:
            language = self.model.languages[0]
        if language is not None and self.model.languages is None:
            language = None
        return language

    def _send_as_file(self, buffer):
        return flask.send_file(
            buffer,
            mimetype='binary/octet-stream'
        )

    def _run_model(self, **kwargs):
        buffer = io.BytesIO()
        self.tts.tts_to_file(file_path=buffer, **kwargs)
        return self._send_as_file(buffer)

    def _run_synthesizer(self, **kwargs):
        buffer = io.BytesIO()
        wav = self.tts.synthesizer.tts(**kwargs)
        self.tts.synthesizer.save_wav(wav=wav, path=buffer, pipe_out=None)
        return self._send_as_file(buffer)

    def dub(self):
        try:
            model = flask.request.json['model']
            self._load_model_internal(model)
            text = flask.request.json['text']
            voice = flask.request.json['voice']
            language = flask.request.json['language']

            language = self._get_language(language)
            if voice is None and self.model.speakers is not None:
                voice = self.model.speakers[0]

            return self._run_synthesizer(text=text, speaker_name=voice, language_name=language)
        except:
            return traceback.format_exc(), 500

    def voice_clone(self):
        try:
            model = flask.request.json['model']
            self._load_model_internal(model)
            text = flask.request.json['text']
            voice = flask.request.json['voice']
            language = flask.request.json['language']
            language = self._get_language(language)
            voice_path = f'/resources/voices/{voice}/'
            voice_paths = [voice_path+file for file in os.listdir(voice_path)]
            if len(voice_paths) == 0:
                raise ValueError(f"Voice {voice} folder doesn't contain any files")
            return self._run_model(text=text, speaker_wav=voice_paths, language=language)
        except:
            return traceback.format_exc(), 500





from brainbox.framework import OnDemandDockerApi, FileLike, FileIO, File, RunConfiguration
from .controller import BoilerplateOnDemandController
from .settings import BoilerplateOnDemandSettings

class BoilerplateOnDemand(OnDemandDockerApi[BoilerplateOnDemandSettings, BoilerplateOnDemandController]):
    def _monitor_progress(self, line: str):
        try:
            if not line.endswith('%'):
                return
            value = int(line[:-1])
            self.context.logger.report_progress(value/100)
        except:
            pass

    def execute(self, input: str):
        FileIO.write_text(input, self.controller.resource_folder()/'input')

        configuration = RunConfiguration(
            detach_and_interactive=False
        )
        self.controller.run_with_configuration(configuration, self._monitor_progress)

        return FileIO.read_text(self.controller.resource_folder()/'output')


    Controller = BoilerplateOnDemandController
    Settings = BoilerplateOnDemandSettings



from typing import *
from dataclasses import dataclass, field
from enum import Enum
from datetime import datetime

@dataclass
class OperatorMessage:
    class Type(Enum):
        accepted = 0
        report_progress = 1
        log = 2
        result = 3
        error = 4



    id: str
    type: 'OperatorMessage.Type'
    payload: Any = None
    timestamp: datetime = field(default_factory=datetime.now)




from typing import Iterable
from unittest import TestCase
from ....framework import (
    RunConfiguration, TestReport, BrainboxImageBuilder, IImageBuilder, DockerWebServiceController,
    BrainBoxApi, BrainBoxTask, FileIO, INotebookableController, IModelDownloadingController, DownloadableModel
)
from .settings import HelloBrainBoxSettings
from .model import HelloBrainBoxModel
from pathlib import Path


class HelloBrainBoxController(
    DockerWebServiceController[HelloBrainBoxSettings],
    INotebookableController,
    IModelDownloadingController
):
    def get_image_builder(self) -> IImageBuilder|None:
        return BrainboxImageBuilder(
            Path(__file__).parent,
            '3.11.11'
        )

    def get_downloadable_model_type(self) -> type[DownloadableModel]:
        return HelloBrainBoxModel

    def get_service_run_configuration(self, parameter: str|None) -> RunConfiguration:
        if parameter is None:
            parameter = 'no_parameter'
        return RunConfiguration(
            parameter,
            publish_ports={self.connection_settings.port:8080},
            command_line_arguments=['--setting', self.settings.setting, '--parameter', parameter],
        )

    def get_notebook_configuration(self) -> RunConfiguration|None:
        return self.get_service_run_configuration('').as_notebook_service()

    def get_default_settings(self):
        return HelloBrainBoxSettings()

    def create_api(self):
        from .api import HelloBrainBox
        return HelloBrainBox()

    def post_install(self):
        FileIO.write_text("HelloBrainBox resource", self.resource_folder()/'resource')
        FileIO.write_text("HelloBrainBox nested resource", self.resource_folder('nested') / 'resource')


    def _self_test_internal(self, api: BrainBoxApi, tc: TestCase) -> Iterable:
        from .api import HelloBrainBox

        api.execute(BrainBoxTask.call(HelloBrainBox, 'test_parameter').json('test_argument_json'))


        yield (
            TestReport
            .last_call(api)
            .href('json')
            .with_comment("Returns JSON as a string. This string is stored in the database")
        )

        api.execute(BrainBoxTask.call(HelloBrainBox, 'test_parameter').file('test_argument_file'))
        yield (
            TestReport
            .last_call(api)
            .href('file')
            .result_is_file()
            .with_comment("Returns a json as a file. It's content is not stored in the database, but in the file cache")
        )

        api.execute(BrainBoxTask.call(HelloBrainBox, 'test_parameter').resources())
        yield (
            TestReport
            .last_call(api)
            .href('resources')
            .with_comment("Returns a json with the list of resources: files that are stored at the server outside of the cache, and are shared with the container")
        )




from .docker_controller import DockerController, TSettings
from abc import ABC, abstractmethod
from .run_configuration import RunConfiguration
from .connection_settings import ConnectionSettings
import requests
from ...common import ApiUtils

class DockerWebServiceController(DockerController[TSettings], ABC):
    @abstractmethod
    def get_service_run_configuration(self, parameter: str|None) -> RunConfiguration:
        pass

    @abstractmethod
    def create_api(self):
        pass

    @property
    def connection_settings(self) -> ConnectionSettings:
        settings = self.settings
        if not hasattr(settings, 'connection'):
            raise ValueError(f"Settings do not have `connection` field defined, impossible to use with {type(self).__name__}")
        connection = settings.connection
        if not isinstance(connection, ConnectionSettings):
            raise ValueError(f"Settings field `connection` is not of ConnectionSettings type")
        return connection

    def find_api(self, instance_id: str):
        api = self.create_api()
        from .docker_web_service_api import DockerWebServiceApi
        if not isinstance(api, DockerWebServiceApi):
            raise ValueError(f"Unexpected type {api} returned, expected an object of DockerWebServiceApi's subclass")
        api.controller = self
        api._container_parameter = self.instance_id_to_parameter(instance_id)
        return api

    @property
    def address(self) -> str:
        return f'127.0.0.1:{self.connection_settings.port}'

    def endpoint(self, endpoint=''):
        return 'http://'+self.address + endpoint

    def wait_for_boot(self):
        ApiUtils.wait_for_reply(
            self.endpoint(),
            self.connection_settings.loading_time_in_seconds,
            self.get_name()
        )

    def is_reachable(self):
        try:
            reply = requests.get(self.endpoint())
        except:
            return False
        if reply.status_code != 200:
            return False
        return True


    def run(self, parameter: str|None = None):
        self.stop_all()
        container_id = self.run_with_configuration(self.get_service_run_configuration(parameter))
        self.wait_for_boot()
        return container_id





import os
from typing import Iterable
from unittest import TestCase
from ....framework import (
    RunConfiguration, TestReport, BrainboxImageBuilder, IImageBuilder, DockerWebServiceController,
    BrainBoxApi, BrainBoxTask, FileIO, INotebookableController, IModelDownloadingController, DownloadableModel,
    File
)
from ...common import VOICEOVER_TEXT, check_if_its_sound
from .settings import ZonosSettings
from pathlib import Path


class ZonosController(
    DockerWebServiceController[ZonosSettings],
    INotebookableController,
):
    def get_image_builder(self) -> IImageBuilder|None:
        return BrainboxImageBuilder(
            Path(__file__).parent,
            'pytorch/pytorch:2.6.0-cuda12.4-cudnn9-devel',
            ('espeak-ng','git','ffmpeg'),
            BrainboxImageBuilder.Repository(
                'https://github.com/Zyphra/Zonos',
                'c6f7704',
            ),
            custom_dependencies=(
                BrainboxImageBuilder.CustomDependencies(
                    ('torch==2.6.0+cu124', 'torchaudio==2.6.0+cu124'),
                    index_url='https://download.pytorch.org/whl/cu124'
                ),
                BrainboxImageBuilder.CustomDependencies(
                    ('mamba-ssm==2.2.4', 'flash_attn==2.7.4.post1', 'causal-conv1d==1.5.0.post8'),
                    no_build_isolation=True,
                    no_binary=('mamba-ssm',),
                ),
                BrainboxImageBuilder.Dependencies()
            ),
            keep_dockerfile = True

        )

    def get_service_run_configuration(self, parameter: str|None) -> RunConfiguration:
        if parameter is not None:
            raise ValueError(f"`parameter` must be None for {self.get_name()}")
        return RunConfiguration(
            publish_ports={self.connection_settings.port:8080},
            mount_resource_folders= {
                'cache': '/home/app/.cache',
                'speakers': '/speakers',
                'voices': '/voices'
            },
        )

    def get_notebook_configuration(self) -> RunConfiguration|None:
        return self.get_service_run_configuration(None).as_notebook_service()

    def get_default_settings(self):
        return ZonosSettings()

    def create_api(self):
        from .api import Zonos
        return Zonos()

    def post_install(self):
        path = self.resource_folder('cache')
        if path.is_dir() and len(os.listdir(path))==2:
            print('Already installed')
            return
        self.run_with_configuration(self.get_service_run_configuration(None).as_service_worker('--install'))

    def _self_test_internal(self, api: BrainBoxApi, tc: TestCase) -> Iterable:
        from .api import Zonos

        speaker_sample = File.read(Path(__file__).parent/'app/lina.mp3')
        api.execute(BrainBoxTask.call(Zonos).train('test_speaker', [speaker_sample]))

        yield (
            TestReport
            .last_call(api)
            .href('train')
        )

        result = api.execute(BrainBoxTask.call(Zonos).voiceover(VOICEOVER_TEXT, 'test_speaker'))
        check_if_its_sound(api.open_file(result).content, tc)
        yield (
            TestReport
            .last_call(api)
            .href('voiceover')
            .result_is_file(File.Kind.Audio)
        )

        result = api.execute(BrainBoxTask.call(Zonos).voiceover(VOICEOVER_TEXT, 'test_speaker', speaking_rate=10))
        check_if_its_sound(api.open_file(result).content, tc)
        yield (
            TestReport
            .last_call(api)
            .href('voiceover')
            .result_is_file(File.Kind.Audio)
            .with_comment("Slower speech")
        )

        result = api.execute(BrainBoxTask.call(Zonos).voiceover(VOICEOVER_TEXT, 'test_speaker', speaking_rate=20))
        check_if_its_sound(api.open_file(result).content, tc)
        yield (
            TestReport
            .last_call(api)
            .href('voiceover')
            .result_is_file(File.Kind.Audio)
            .with_comment("Faster speech")
        )


        result = api.execute(BrainBoxTask.call(Zonos).voiceover(VOICEOVER_TEXT, 'test_speaker', emotion=Zonos.Emotions.Happiness))
        check_if_its_sound(api.open_file(result).content, tc)
        yield (
            TestReport
            .last_call(api)
            .href('voiceover')
            .result_is_file(File.Kind.Audio)
            .with_comment("With emotion")
        )

CUDA_RUNTIME_INSTALLATION = '''
RUN apt-get update && apt-get install -y --no-install-recommends \
      curl ca-certificates git ffmpeg espeak-ng build-essential ninja-build cmake && \
    curl -fsSL -o /tmp/cuda-keyring.deb https://developer.download.nvidia.com/compute/cuda/repos/debian12/x86_64/cuda-keyring_1.1-1_all.deb && \
    dpkg -i /tmp/cuda-keyring.deb && rm -f /tmp/cuda-keyring.deb && \
    apt-get update && apt-get install -y --no-install-recommends cuda-toolkit-12-4 && \
    rm -rf /var/lib/apt/lists/*
'''.strip()


from typing import Iterable
from unittest import TestCase
from ....framework import (
    RunConfiguration, TestReport, BrainboxImageBuilder, IImageBuilder, DockerWebServiceController,
    BrainBoxApi, BrainBoxTask, INotebookableController, File, SmallImageBuilder
)
from .settings import InsightFaceSettings
from pathlib import Path
import os

class InsightFaceController(
    DockerWebServiceController[InsightFaceSettings],
    INotebookableController
):
    def get_image_builder(self) -> IImageBuilder|None:
        return BrainboxImageBuilder(
            Path(__file__).parent,
            '3.11.11',
            allow_arm64=True
        )


    def get_service_run_configuration(self, parameter: str|None) -> RunConfiguration:
        if parameter is not None:
            raise ValueError(f"`parameter` must be None for {self.get_name()}")
        config = RunConfiguration(
            publish_ports={self.settings.connection.port: 8084},
        )
        return config

    def get_notebook_configuration(self) -> RunConfiguration|None:
        return self.get_service_run_configuration(None).as_notebook_service()

    def get_default_settings(self):
        return InsightFaceSettings()

    def create_api(self):
        from .api import InsightFace
        return InsightFace()

    def post_install(self):
        pass


    def _self_test_internal(self, api: BrainBoxApi, tc: TestCase) -> Iterable:
        from .api import InsightFace
        file = File.read(Path(__file__).parent / 'files/image.png')


        result = api.execute(BrainBoxTask.call(InsightFace).analyze(image=file))
        yield TestReport.last_call(api)

        tc.assertEqual(5, len(result))

